#include "spatial.h"
#include "../../functions/functions.h"

namespace Filtering { namespace AnisoTools {


template<int w, int h> int sad(const unsigned char *pSrc, int nSrcPitch, const unsigned char *pDst, int nDstPitch)
{
   int s = 0;

   for ( int y = 0; y < h; y++, pSrc += nSrcPitch, pDst += nDstPitch )
      for ( int x = 0; x < w; x++ )
         s += abs<int>( pSrc[ x ] - pDst[ x ] );
         
   return s;
}

static inline int median(int x, int y, int z)
{
   if ( x < y )
      if ( z < x )
         return x;
      else if ( z < y )
         return z;
      else
         return y;
   else
      if ( z > x )
         return x;
      else if ( z > y )
         return z;
      else
         return y;
}

#define TEST_MV(_x, _y) \
   do { \
      int __x = clip<int>(_x, minx, maxx); \
      int __y = clip<int>(_y, miny, maxy); \
      int s = sad<16, 16>(pSrc, nSrcPitch, pDst + (__x) + (__y) * nDstPitch, nDstPitch) + lambda * (abs<int>( (__x) - predictors[nPredictors-1].x ) + abs<int>( (__y) - predictors[nPredictors-1].y )); \
      if ( s < best.sad ) \
      { \
         best.mv.x = __x; \
         best.mv.y = __y; \
         best.sad = s; \
      } \
   } while( 0 )

void memb(MB *mb, int x, int y, int w, int h, const unsigned char *pSrc, int nSrcPitch, const unsigned char *pDst, int nDstPitch, int minx, int maxx, int miny, int maxy)
{
   const int lambda = 10;
   MB best = { { 0, 0 }, 16576000 };
   MV predictors[10];
   int nPredictors = 0;

   /* predictors */
   if ( x > 0 )
   {
      predictors[nPredictors++] = mb[-1].mv;
      if ( y > 0 )
         predictors[nPredictors++] = mb[-1-w].mv;
   }
   if ( y > 0 )
   {
      predictors[nPredictors++] = mb[-w].mv;
      if ( x < w - 1 )
         predictors[nPredictors++] = mb[1-w].mv;
   }
   if (!nPredictors)
   {
      predictors[nPredictors].x = 0;
      predictors[nPredictors++].y = 0;
   }
   else if (nPredictors > 2)
   {
      predictors[nPredictors].x = median( predictors[0].x, predictors[1].x, predictors[2].x );
      predictors[nPredictors++].y = median( predictors[0].y, predictors[1].y, predictors[2].y );
   }

   for ( int i = 0; i < nPredictors; i++ )
      TEST_MV( predictors[i].x, predictors[i].y );

   while ( true )
   {
      static const int pos[] = { -1, 0, 0, -1, 1, 0, 0, 1, };
      MB curr = best;
      for ( int i = 0; i < sizeof(pos) / (2 * sizeof(pos[0])); i++ )
         TEST_MV( curr.mv.x + pos[2*i], curr.mv.y + pos[2*i+1] );
      if ( best.sad == curr.sad )
         break;
   }
   *mb = best;
}

void me(MB *mb, int w, int h, const unsigned char *pSrc, int nSrcPitch, const unsigned char *pDst, int nDstPitch, int paddw, int paddh)
{
   for ( int y = 0; y < h; y++ )
      for ( int x = 0; x < w; x++ )
         memb( &mb[x + y * w], x, y, w, h, pSrc + x * 16 + y * 16 * nSrcPitch, nSrcPitch, pDst + x * 16 + y * 16 * nDstPitch, nDstPitch, -x * 16, (w - 1 - x) * 16, -y * 16, (h - 1 - y) * 16);
}

static void filter_plane(const Plane<PIXTYPE> &destination, const Plane<const PIXTYPE> &source,
                         const Plane<const PIXTYPE> &forward, const Plane<const PIXTYPE> &backward,
                         const Plane<const PIXTYPE> &gradient, const int *spatial, const int *temporal,
                         int sstep, int tstep, const MB *mbf, const MB *mbb, int sx, int sy)
{
   const int bsx = 16 >> sx;
   const int bsy = 16 >> sy;
   for ( int y = 0; y < destination.height(); y++ )
   {
      for ( int x = 0; x < destination.width(); x++ )
      {
         const int c = source[ x + y * source.pitch() ];
         int sval = 0, tval = 0;

         if ( spatial )
         {
            const int cg = gradient[ x + y * gradient.pitch() ];

            sval += (spatial[ gradient[ x - 1 + y * gradient.pitch() ] - cg ] * (source[ x - 1 + y * source.pitch() ] - c) + (1 << (SPRECISION - 1))) >> SPRECISION;
            sval += (spatial[ gradient[ x + 1 + y * gradient.pitch() ] - cg ] * (source[ x + 1 + y * source.pitch() ] - c) + (1 << (SPRECISION - 1))) >> SPRECISION;
            sval += (spatial[ gradient[ x + (y + 1) * gradient.pitch() ] - cg ] * (source[ x + (y + 1) * source.pitch() ] - c) + (1 << (SPRECISION - 1))) >> SPRECISION;
            sval += (spatial[ gradient[ x + (y - 1) * gradient.pitch() ] - cg ] * (source[ x + (y - 1) * source.pitch() ] - c) + (1 << (SPRECISION - 1))) >> SPRECISION;

            sval = (sval * sstep + (2 << STPRECISION)) >> (2 + STPRECISION);
         }
         if ( temporal )
         {
            const int xf = mbf[ (x / bsx) + (y / bsy) * (source.width() / bsx) ].mv.x >> sx;
            const int yf = mbf[ (x / bsx) + (y / bsy) * (source.width() / bsx) ].mv.y >> sy;

            const int xb = mbb[ (x / bsx) + (y / bsy) * (source.width() / bsx) ].mv.x >> sx;
            const int yb = mbb[ (x / bsx) + (y / bsy) * (source.width() / bsx) ].mv.y >> sy;

            tval += temporal[ forward[ x + xf + (y + yf) * forward.pitch() ] - c ];
            tval += temporal[ backward[ x + xb + (y + yb) * backward.pitch() ] - c ];

            tval = (tval * tstep + (1 << STPRECISION)) >> (1 + STPRECISION);
         }

         destination[ x + y * destination.pitch() ] = c + sval + tval;
      }
   }
}

static void filter_gradient(const Plane<PIXTYPE> &gradient, const Plane<const PIXTYPE> &source, int nCenter)
{
   for ( int y = 0; y < gradient.height(); y++ )
   {
      for ( int x = 0; x < gradient.width(); x++ )
      {
         int val = 0;

         for (int i = -1; i < 2; i++ )
            for (int j = -1; j < 2; j++ )
               val += (i == 0 && j == 0 ? nCenter : 1) * source[x + i + (y + j) * source.pitch()];

         gradient[x + y * gradient.pitch()] = (val + (nCenter + 8) / 2) / (nCenter + 8);
      }
   }
}

Frame<Byte> Filters::Spatial::Filter::get_frame(int n, const Frame<Byte> &output_frame)
{
   int start = nLastFrame == n - 1 ? nIterations : -1;

   for ( int k = start; k <= nIterations; k++ )
   {
      if ( k > 0 )
      {
         me( mbs[0][k-1], nWidth / 16, nHeight / 16, 
            childs[0]->get_const_frame( n + k - 1 ).plane(0), childs[0]->get_const_frame( n + k - 1 ).plane(0).pitch(),
            childs[0]->get_const_frame( n + k - 2 ).plane(0), childs[0]->get_const_frame( n + k - 2 ).plane(0).pitch(),
            0, 0 );
         me( mbs[1][k-1], nWidth / 16, nHeight / 16, 
            childs[0]->get_const_frame( n + k - 1 ).plane(0), childs[0]->get_const_frame( n + k - 1 ).plane(0).pitch(),
            childs[0]->get_const_frame( n + k ).plane(0), childs[0]->get_const_frame( n + k ).plane(0).pitch(),
            0, 0 );
      }

      Functions::ConvertFrame<PIXTYPE, Byte>( frames[ k ]->select(1, 1), childs[0]->get_const_frame( n + k ), PRECISION );

      for ( int j = k - 1; j >= 0; j-- )
      {
         Functions::PaddFrame<PIXTYPE>( *frames[j], 1, 1 );
         for ( int i = 0; i < plane_counts[C]; i++ )
         {
            filter_gradient( gradient.select(1, 1).plane(i), frames[j]->select(1, 1).plane(i), nCenter );
            Functions::PaddFrame<PIXTYPE>( gradient, 1, 1 );
            filter_plane( frame->select(1, 1).plane(i), frames[j]->select(1, 1).plane(i),
                          frames[j-1]->select(1, 1).plane(i), frames[j+1]->select(1, 1).plane(i),
                          gradient.select(1, 1).plane(i), pSpatials[i], pTemporals[i], nSSteps[i], nTSteps[i],
                          mbs[0][j], mbs[1][j], i ? 1 : 0, i ? 1 : 0 );
         }
         Frame<PIXTYPE> *p = frame;
         frame = frames[j];
         frames[j] = p;
      }
   }

   Functions::ConvertFrame<Byte, PIXTYPE>( output_frame, Frame<const PIXTYPE>(frames[0]->select(1, 1)), -PRECISION );

   nLastFrame = n;

   Frame<PIXTYPE> *p = frame;
   frame = frames[-1];
   memmove(&frames[-1], &frames[0], sizeof(frames[0])*(nIterations+1));
   frames[nIterations] = p;

   if (nIterations > 1)
   {
      MB *mb = mbs[0][0];
      memmove(&mbs[0][0], &mbs[0][1], sizeof(mbs[0][0])*(nIterations-1));
      mbs[0][nIterations-1] = mb;

      mb = mbs[1][0];
      memmove(&mbs[1][0], &mbs[1][1], sizeof(mbs[1][0])*(nIterations-1));
      mbs[1][nIterations-1] = mb;
   }

   return output_frame;
}

} } // namespaces AnisoTools, Filtering
