#ifndef __At_Spatial_H__
#define __At_Spatial_H__

#include "../../../common/functions/functions.h"
#include "../../../common/filter/filter.h"

namespace Filtering { namespace AnisoTools {
   
typedef struct MV {
   int x, y;
} MV;

typedef struct MB { 
   MV mv;
   int sad;
} MB;
   
namespace Filters { namespace Spatial {

#define PIXTYPE Short
#define PRECISION 7
#define SPRECISION 15
#define STPRECISION 8


class Filter : public ChildFilter {

   PIXTYPE *pBuffer;
   Frame<PIXTYPE> **frames;
   Frame<PIXTYPE> *frame;
   Frame<PIXTYPE> gradient;

   MB **mbs[2];

   int *pSpatials[3];
   int *pTemporals[3];
   int nSSteps[3];
   int nTSteps[3];
   int nIterations;
   int nCenter;

   int nLastFrame;

public:
   Filter(const Parameters &parameters) : ChildFilter(parameters)
   {
      frames = NULL;
      pBuffer = NULL;
      memset( pSpatials, 0, sizeof( pSpatials ) );
      memset( pTemporals, 0, sizeof( pTemporals ) );
      mbs[0] = mbs[1] = NULL;

      nLastFrame = -2;

      nIterations = parameters["iterations"].toInt();
      if ( nIterations < 0 )
      {
         error = "iteration must be superior or equal to 0";
         return;
      }

      nCenter = parameters["center"].toInt();
      if ( nCenter < 1 )
      {
         error = "center must be superior or equal to 1";
         return;
      }

      frames = new Frame<PIXTYPE> *[ nIterations + 2 ];
      PIXTYPE *p = pBuffer = new PIXTYPE[(nIterations + 4) * (nWidth + 2) * (nHeight + 2) * 3];
      gradient = Frame<PIXTYPE>(&p, nWidth, nHeight, C, 1, 1);
      for ( int i = 0; i < nIterations + 2; i++ )
         frames[i] = new Frame<PIXTYPE>(&p, nWidth, nHeight, C, 1, 1);
      frame = new Frame<PIXTYPE>(&p, nWidth, nHeight, C, 1, 1);

      mbs[0] = new MB*[ nIterations + 2 ];
      mbs[1] = new MB*[ nIterations + 2 ];
      for ( int i = 0; i < nIterations + 2; i++ )
      {
         mbs[0][i] = new MB[ (nWidth / 16) * (nHeight / 16) ];
         mbs[1][i] = new MB[ (nWidth / 16) * (nHeight / 16) ];
      }

      frames += 1;

      for ( int i = 0; i < 3; i++ )
      {
         double spatial = parameters["edge"].toFloat() * (1 << PRECISION);
         double temporal = parameters["motion"].toFloat() * (1 << PRECISION);

         pSpatials[i] = (new int[512 << PRECISION]) + (256 << PRECISION);
         pTemporals[i] = (new int[512 << PRECISION]) + (256 << PRECISION);

         for ( int j = -(256 << PRECISION); j < (256 << PRECISION); j++ )
         {
            pSpatials[i][j] = convert<int, Double>((1 << SPRECISION) * exp( -abs<Double>( j / spatial ) ));
            pTemporals[i][j] = convert<int, Double>(j * exp( -abs<Double>( j / temporal ) ));
         }
      }

      nSSteps[0] = convert<int, Double>((1 << STPRECISION) * parameters["sstep"].toFloat());
      nSSteps[1] = convert<int, Double>((1 << STPRECISION) * parameters["sstep"].toFloat());
      nSSteps[2] = convert<int, Double>((1 << STPRECISION) * parameters["sstep"].toFloat());

      nTSteps[0] = convert<int, Double>((1 << STPRECISION) * parameters["tstep"].toFloat());
      nTSteps[1] = convert<int, Double>((1 << STPRECISION) * parameters["tstep"].toFloat());
      nTSteps[2] = convert<int, Double>((1 << STPRECISION) * parameters["tstep"].toFloat());
   }
   ~Filter()
   {
      if ( pBuffer ) delete[] pBuffer;
      if ( frame ) delete frame;
      if ( frames )
      {
         for ( int i = -1; i <= nIterations; i++ )
            delete frames[i];
         delete[] (frames - 1);
      }

      if ( mbs[0] ) 
      {
         for ( int i = 0; i < nIterations + 2; i++ )
         {
            delete[] mbs[0][i];
            delete[] mbs[1][i];
         }
         delete[] mbs[0];
         delete[] mbs[1];
      }
      
      for ( int i = 0; i < 3; i++ )
      {
         if ( pSpatials[i] ) delete[] (pSpatials[i] - (256 << PRECISION));
         if ( pTemporals[i] ) delete[] (pTemporals[i] - (256 << PRECISION));
      }
   }

   virtual Frame<Byte> get_frame(int n, const Frame<Byte> &output_frame);

   static Signature filter_signature()
   {
      Signature signature = "at_denoise";

      signature.add( Parameter( TYPE_CLIP, "" ) );
      signature.add( Parameter( 0.8f, "tstep" ) );
      signature.add( Parameter( 0.8f, "sstep" ) );
      signature.add( Parameter( 1.0f, "edge" ) );
      signature.add( Parameter( 1.0f, "motion" ) );
      signature.add( Parameter( 5, "iterations" ) );
      signature.add( Parameter( 4, "center" ) );

      return signature;
   }

};

} } } } // namespaces Spatial, Filters, AnisoTools, Filtering

#endif // __At_Spatial_H__